"""
# Get an implant with suggested targets from a dict
>>> shield = npc_shields.shields.get_shield('2002')
>>> suggested_targets = npc_shields.insertions.InsertionRecord(
...     shield=shield,
...     probes={
...         'A': 'A1',
...         'B': None,
...         'C': 'C1',
...         'D': 'D2',
...         'E': 'F1',
...         'F': 'F2',
...     },
...     session='366122_20231201',
...     experiment_day=1,
... )
>>> w = InsertionWidget(
...     insertion=suggested_targets,
...     save_paths='example-record.json'
... ) # doctest: +SKIP

# # Get an implant with suggested targets from a previously-saved record
# >>> suggested_targets = InsertionWidget.from_record(pathlib.Path('examples/example-record.json'))
"""
from __future__ import annotations

import datetime
import functools
import json
import pathlib
from collections.abc import Iterable
from typing import Any

import IPython.display
import ipywidgets as ipw
import npc_session
import pydantic
from typing_extensions import Self

import npc_shields.injections
import npc_shields.insertions
import npc_shields.shields
import npc_shields.types


class InsertionWidget(ipw.HBox):
    """Displays implant drawing, configurable probe-hole assignments, and buttons for interaction"""

    @classmethod
    def from_record(cls, json_path: pathlib.Path, **kwargs) -> Self:
        """Load widget from an existing insertion record."""
        return cls(
            insertion=npc_shields.insertions.InsertionRecord.from_json(
                json.loads(json_path.read_text())
            ),
            save_paths=json_path,
            **kwargs,
        )

    def __init__(
        self,
        insertion: npc_shields.types.Insertion,
        save_paths: str | pathlib.Path | Iterable[pathlib.Path],
        read_only: bool = False,
        **hbox_kwargs,
    ) -> None:
        for k, v in hbox_kwargs.items():
            setattr(self, k, v)

        if isinstance(save_paths, str):
            save_paths = pathlib.Path(save_paths)
        if not isinstance(save_paths, Iterable):
            save_paths = (save_paths,)
        self.save_paths: Iterable[pathlib.Path] = save_paths  # type: ignore [assignment]

        self.insertion = insertion
        self.initial_targets = dict(insertion.probes)
        self.probe_letters = sorted(self.insertion.probes.keys())

        self.probe_hole_sliders = [
            ipw.SelectionSlider(
                options=["none", *self.insertion.shield.labels],
                value=self.insertion.probes[probe] or "none",
                description=f"probe {probe}",
                continuous_update=True,
                orientation="horizontal",
                readout=True,
            )
            for probe in self.probe_letters
        ]

        def update_insertions(**kwargs) -> None:
            "Update probe-hole assignments when sliders are changed"
            assignments = {k: v if v != "none" else None for k, v in kwargs.items()}
            self.insertion.probes.update(assignments)
            self.update_display()

        self.interactive_insertion = ipw.interactive_output(
            f=update_insertions,
            controls=dict(zip(self.probe_letters, self.probe_hole_sliders)),
        )

        # additional ui elements --------------------------------------------------------------- #
        self.note_entry_boxes = [
            ipw.Text(
                value=self.insertion.notes[probe],
                placeholder=f"Add notes for probe {probe}",
                continuous_update=True,
            )
            for probe in self.probe_letters
        ]
        "Text entry box for notes for each probe"

        self.slider_ui = (
            ipw.VBox([*self.probe_hole_sliders])
            if (self.probe_hole_sliders[0].orientation == "horizontal")
            else ipw.HBox([*self.probe_hole_sliders])
        )
        self.notes_ui = ipw.VBox([*self.note_entry_boxes])
        # -------------------------------------------------------------------------------------- #
        self.slider_notes_ui = ipw.HBox([self.slider_ui, self.notes_ui])
        # -------------------------------------------------------------------------------------- #

        self.save_button = ipw.Button(description="Save", button_style="success")
        self.clear_button = ipw.Button(description="Clear", button_style="warning")
        self.reload_button = ipw.Button(
            description="Reload targets", button_style="info"
        )
        self.save_button.on_click(functools.partial(self.save_button_clicked, self))
        self.clear_button.on_click(functools.partial(self.clear_button_clicked, self))
        self.reload_button.on_click(functools.partial(self.reload_button_clicked, self))
        # -------------------------------------------------------------------------------------- #
        self.button_ui = ipw.HBox(
            [self.clear_button, self.reload_button, self.save_button]
        )
        # -------------------------------------------------------------------------------------- #

        self.output = ipw.Output()
        "Console for displaying messages"

        self.console_clear()

        left_box = self.interactive_insertion
        right_box = ipw.VBox([self.slider_notes_ui, self.button_ui, self.output])
        super().__init__(
            [
                left_box,
                right_box,
            ]
        )
        "Feed all UI elements into superclass widget"

        self.layout = ipw.Layout(width="100%")

        # UI adjustments
        inputs = [
            *self.button_ui.children,
            *self.probe_hole_sliders,
            *self.note_entry_boxes,
        ]
        if read_only:
            self.console_print("Insertion record loaded (read-only).")
            for input in inputs:
                input.disabled = True
                if isinstance(input, ipw.Button):
                    input.button_style = ""

    # end of init - widget returned/displayed ----------------------------------------------- #

    @property
    def probe_hole_assignments_display_handle(self) -> IPython.display.DisplayHandle:
        if not hasattr(self, "_probe_hole_assignments_display"):
            self._probe_hole_assignments_display = IPython.display.DisplayHandle()
        return self._probe_hole_assignments_display

    def update_display(self) -> None:
        self.probe_hole_assignments_display_handle.display(
            ipw.HTML(
                self.insertion.to_svg(),
                layout=ipw.Layout(align_content="center", object_fit="scale-down"),
                # layout not working
            )
        )

    def console_print(self, msg: str) -> None:
        with self.output:
            print(f"{datetime.datetime.now().strftime('%H:%M:%S')} {msg}")

    def console_clear(self) -> None:
        msg = " " * 30
        with self.output:
            print(f"{msg}")

    def save_button_clicked(self, *args, **kwargs) -> None:
        for probe, text in zip(self.insertion.probes, self.note_entry_boxes):
            self.insertion.notes[probe] = text.value or None
        text = json.dumps(self.insertion.to_json(), indent=2)
        for path in self.save_paths:
            path.write_text(text)
        self.console_print("Insertions saved.")

    def clear_button_clicked(self, *args, **kwargs) -> None:
        for slider in self.probe_hole_sliders:
            slider.value = "none"
        self.console_clear()

    def reload_button_clicked(self, *args, **kwargs) -> None:
        for probe, slider in zip(self.probe_letters, self.probe_hole_sliders):
            hole = self.initial_targets[probe]
            if hole is not None:
                slider.value = hole
            else:
                slider.value = "none"
        self.console_print("Targets reloaded.")


def get_insertion_widget(
    shield_name: str,
    session: str | npc_session.SessionRecord,
    experiment_day: int,
    save_paths: str | pathlib.Path | Iterable[pathlib.Path],
) -> InsertionWidget:
    # """

    # >>> w = get_insertion_widget(
    # ...     shield_name='2002',
    # ...     session='366122_20231201',
    # ...     experiment_day=1,
    # ...     save_paths=pathlib.Path('examples/example-record.json'),
    # ... )
    # """
    return InsertionWidget(
        insertion=npc_shields.insertions.InsertionRecord(
            shield=npc_shields.shields.get_shield(shield_name),
            session=session,
            experiment_day=experiment_day,
            probes=None,
            notes=None,
        ),
        save_paths=save_paths,
    )


class InjectionWidget(ipw.VBox):
    """
    A widget for a single session, allowing one or more injections to be added.
    """

    injections: set[npc_shields.types.Injection]
    gridspec_kwargs: dict[str, Any]
    shield: npc_shields.types.Shield | None

    def __init__(
        self,
        session: str | npc_session.SessionRecord,
        injection_day: int,
        save_paths: str | pathlib.Path | Iterable[pathlib.Path],
        shield_name: str | None = None,
        injection_cls: type[pydantic.BaseModel] = npc_shields.injections.Injection,
        **vbox_kwargs,
    ) -> None:
        self.session = session
        self.injection_day = injection_day
        self.save_paths = save_paths  # type: ignore [assignment]
        if shield_name is not None:
            self.shield = npc_shields.shields.get_shield(shield_name)
        else:
            self.shield = None
        self.injection_cls = injection_cls
        self.injections = self.load_existing_injections()

        def get_hint(name, field) -> str:
            if name == "start_time":
                return "[required] YYYY-MM-DD HH:MM"
            return f"{'[required]' if field.is_required() else ''}"

        self.text_entry_boxes = {
            name: ipw.Text(
                description=name,
                placeholder=get_hint(name, field),
                tooltip=field.description or name,
                continuous_update=True,
                layout=ipw.Layout(width="100%"),
            )
            for name, field in self.injection_cls.model_fields.items()
            if name not in ("shield")
        }
        self._apply_default_injection_values()
        self.text_entry_grid = ipw.GridBox(
            [v for v in self.text_entry_boxes.values()],
        )
        hbox_elements = [self.text_entry_grid]
        if self.shield:
            hbox_elements.append(
                ipw.HTML(
                    self.shield.svg.read_text(),
                    layout=ipw.Layout(align_content="center", object_fit="scale-down"),
                    # layout not working
                )
            )
        hbox = ipw.HBox(hbox_elements)

        self.add_injection_button = ipw.Button(
            description="Add this injection",
            button_style="success",
            layout=ipw.Layout(width="30%"),
            tooltip="Append the current set of parameters as a unique injection for this session",
        )
        self.add_injection_button.on_click(lambda _: self.add_injection())
        self.console = ipw.Output()

        if self.injections:
            with self.console:
                print(
                    f"Loaded existing injections [total: {len(self.injections)} injections]"
                )
        super().__init__([hbox, self.add_injection_button, self.console], **vbox_kwargs)

    @property
    def save_paths(self) -> tuple[pathlib.Path, ...]:
        return self._save_paths

    @save_paths.setter
    def save_paths(self, paths: str | pathlib.Path | Iterable[pathlib.Path]) -> None:
        if isinstance(paths, str):
            paths = (pathlib.Path(paths),)
        if not isinstance(paths, Iterable):
            paths = (paths,)
        paths = (pathlib.Path(path) for path in paths)
        self._save_paths = tuple(
            path / "injections.json" if path.is_dir() else path for path in paths
        )

    def _apply_default_injection_values(self) -> None:
        for name, field in self.injection_cls.model_fields.items():
            if name in ("shield"):
                continue
            if "PydanticUndefined" in field.default.__class__.__name__:
                self.text_entry_boxes[name].value = ""
            elif field.default is None:
                self.text_entry_boxes[name].value = ""
            else:
                self.text_entry_boxes[name].value = str(getattr(field, "default", ""))

    def add_injection(self) -> None:
        try:
            injection = self.injection_cls(
                shield=self.shield,
                **{
                    name: box.value if box.value != "" else None
                    for name, box in self.text_entry_boxes.items()
                },
            )
        except pydantic.ValidationError as e:
            with self.console:
                self.console.clear_output()
                print(f"Error: {e!r}")
            return
        current_num_injections = len(self.injections)
        self.injections.add(injection)  # type: ignore [arg-type]
        self.write_record()
        with self.console:
            self.console.clear_output()
            if current_num_injections == len(self.injections):
                print(
                    f"Injection already exists for these parameters [total: {len(self.injections)} injections]"
                )
            else:
                print(f"Added injection [new total: {len(self.injections)} injections]")

    def load_existing_injections(self) -> set[npc_shields.types.Injection]:
        existing = set()
        for path in self.save_paths:
            if path.exists():
                data = json.loads(path.read_text())
                for injection in data["injections"]:
                    injection["shield"] = npc_shields.shields.get_shield(
                        injection["shield"]["name"]
                    )
                    existing.add(self.injection_cls(**injection))
        return existing  # type: ignore [return-value]

    def create_record(self) -> npc_shields.injections.InjectionRecord:
        return npc_shields.injections.InjectionRecord(
            injections=self.injections,
            session=self.session,
            injection_day=self.injection_day,
        )

    def write_record(self) -> None:
        record = self.create_record()
        for path in self.save_paths:
            path.write_text(json.dumps(record.to_json(), indent=4))


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.FAIL_FAST)



"""
Protocols and type aliases for implants and related models, for type-checking (MyPy).
"""

from __future__ import annotations

import datetime
import pathlib
import typing
from collections.abc import Iterable, MutableMapping
from typing import Any, Literal, Protocol, Union

import npc_session
from typing_extensions import TypeAlias

InsertionProbeMap: TypeAlias = MutableMapping[str, Union[str, None]]
"""Mapping of probe letter to insertion hole in shield, or None if not inserted.
e.g `{"A": "A1", "B": "B2", "C": None, "D": "E2", "E": "E1", "F": "F1"}`
"""


@typing.runtime_checkable
class Shield(Protocol):
    """A specific implant with a diagram and labelled holes"""

    @property
    def name(self) -> str:
        """Colloquial name for the shield, e.g. '2002', 'Templeton'"""
        ...

    @property
    def drawing_id(self) -> int | str:
        """ID of drawing, e.g. from MPE: '0283-200-002'"""
        ...

    @property
    def labels(self) -> Iterable[str]:
        """Original labels for each hole: e.g. A1, A2, A3, B1, B2, B3, B4, etc."""
        ...

    @property
    def svg(self) -> pathlib.Path:
        """Path to SVG diagram of the shield, with labelled holes as text elements"""
        ...

    def __hash__(self) -> int:
        ...

    def to_json(self) -> dict[str, Any]:
        """Get a JSON-serializable representation of the shield."""
        ...


class Insertion(Protocol):
    """A set of probes inserted (or planned to be inserted) into a shield."""

    @property
    def shield(self) -> Shield:
        """The shield that this probe group was or will be inserted into."""
        ...

    @property
    def probes(self) -> InsertionProbeMap:
        """A record of which probes were inserted into which holes in the shield,
        or a set of targets for planned insertions."""
        ...

    @property
    def notes(self) -> dict[str | npc_session.ProbeRecord, str | None]:
        """Text notes for each probe's insertion."""
        ...

    def to_svg(self) -> str:
        """Get the SVG diagram of the shield with inserted probes labelled."""
        ...

    def to_json(self) -> dict[str, Any]:
        """Get a JSON-serializable representation of the insertion."""
        ...


class InsertionRecord(Insertion, Protocol):
    """A record of a set of probes inserted into a shield."""

    @property
    def session(self) -> npc_session.SessionRecord:
        """Record of the session, including subject, date, session index."""
        ...

    @property
    def experiment_day(self) -> int:
        """1-indexed day of experiment for the subject specified in `session`."""
        ...


class Injection(Protocol):
    """An injection through a hole in a shield at a particular brain location (site + depth).

    - should allow for no shield (e.g. burr hole)
    - should record hemisphere
    - may consist of multiple individual injections
    """

    @property
    def shield(self) -> Shield | None:
        """The shield through which the injection was made."""
        ...

    @property
    def shield_hole(self) -> str | None:
        """The hole in the shield through which the injection was made (e.g. 'C3')."""
        ...

    @property
    def location_ap(self) -> float | None:
        """Distance in millimeters from bregma to injection site along
        anterior-posterior axis (+ve is anterior)."""
        ...

    @property
    def location_ml(self) -> float | None:
        """Distance in millimeters from brain midline to injection site along
        medial-lateral axis."""
        ...

    @property
    def target_structure(self) -> str:
        """The intended brain structure for the injection ('VISp' etc.)."""
        ...

    @property
    def hemisphere(self) -> Literal["left", "right"]:
        """The hemisphere of the brain where the injection was made (e.g. 'left', 'right')."""
        ...

    @property
    def depth_um(self) -> float:
        """Depth of the injection, in microns from brain surface."""
        ...

    @property
    def fluorescence_nm(self) -> int | None:
        """Wavelength of fluorescence for the injection."""
        ...

    @property
    def manufacturer(self) -> str | None:
        """Manufacturer of the injected substance."""
        ...

    @property
    def substance(self) -> str:
        """Name of the injected substance."""
        ...

    @property
    def identifier(self) -> str | None:
        """Identifier of the injected substance (e.g. manufacture serial number)."""
        ...

    @property
    def concentration_mg_ml(self) -> float | None:
        """Concentration of the injected substance in milligrams per milliliter."""
        ...

    @property
    def flow_rate_nl_s(self) -> float:
        """Flow rate of the injection in nanoliters per second."""
        ...

    @property
    def total_volume_nl(self) -> float:
        """Total volume injected, in nanoliters."""
        ...

    @property
    def start_time(self) -> datetime.datetime:
        """Time of the first injection, as a datetime object."""
        ...

    @property
    def is_anaesthetized(self) -> bool:
        """Whether the subject was anaesthetized during the injection."""
        ...

    @property
    def is_control(self) -> bool:
        """Whether the purpose of the injection was a control."""
        ...

    @property
    def notes(self) -> str | None:
        """Text notes for the injection."""
        ...

    def to_json(self) -> dict[str, Any]:
        """Get a JSON-serializable representation of the injection."""
        ...


class InjectionRecord(Protocol):
    """A record of a set of injections in a session."""

    @property
    def injections(self) -> Iterable[Injection]:
        """A record of each injection made."""
        ...

    @property
    def session(self) -> npc_session.SessionRecord:
        """Record of the session, including subject, date, session index."""
        ...

    @property
    def experiment_day(self) -> int:
        """1-indexed day of experiment for the subject specified in `session`."""
        ...

    def to_json(self) -> dict[str, Any]:
        """Get a JSON-serializable representation of the injections in a session."""
        ...



from __future__ import annotations

import dataclasses
import datetime
from collections.abc import Iterable
from typing import Any, Literal

import npc_session
import pydantic

import npc_shields.shields
import npc_shields.types


class Injection(pydantic.BaseModel):
    """An injection through a hole in a shield at a particular brain location (site + depth).

    - should allow for no shield (e.g. burr hole)
    - should record hemisphere
    - may consist of multiple individual injections

    >>> i = Injection(
    ...     shield=npc_shields.shields.DR2002,
    ...     shield_hole="D1",
    ...     target_structure="VISp",
    ...     hemisphere="left",
    ...     depth_um=200,
    ...     substance="Fluorogold",
    ...     manufacturer="Sigma",
    ...     identifier="12345",
    ...     total_volume_nl=1.0,
    ...     concentration_mg_ml=10.0,
    ...     flow_rate_nl_s=0.1,
    ...     start_time=datetime.datetime(2023, 1, 1, 12, 0),
    ...     fluorescence_nm=500,
    ...     notes="This was a test injection",
    ...     is_control=False,
    ...     is_anaesthetized=True,
    ... )
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        frozen=True,
    )

    shield: npc_shields.types.Shield | None = None
    """The shield through which the injection was made."""

    @pydantic.field_serializer("shield", when_used="always")
    def serialize_shield_field(
        self, shield: npc_shields.types.Shield
    ) -> dict[str, Any]:
        return shield.to_json() if shield is not None else None

    shield_hole: str | None = None
    """The hole in the shield through which the injection was made (e.g. 'C3')."""

    target_structure: str
    """The intended brain structure for the injection ('VISp' etc.)."""

    hemisphere: Literal["left", "right"] = "left"
    """The hemisphere of the brain where the injection was made ('left' or 'right')."""

    depth_um: float
    """Depth of the injection, in microns from brain surface."""

    location_ap: float | None = None
    """Distance in millimeters from bregma to injection site along
    anterior-posterior axis (+ve is anterior)."""

    location_ml: float | None = None
    """Distance in millimeters from brain midline to injection site along
    medial-lateral axis."""

    substance: str = "muscimol"
    """Name of the injected substance."""

    fluorescence_nm: int | None = None
    """Emission wavelength of the substance injected, if it fluoresces."""

    manufacturer: str | None = "Sigma-Aldrich"
    """Manufacturer of the injected substance."""

    identifier: str | None = "M1523"
    """Identifier of the injected substance (e.g. manufacture serial number)."""

    total_volume_nl: float
    """Total volume injected, in nanoliters."""

    concentration_mg_ml: float | None
    """Concentration of the injected substance in milligrams per milliliter."""

    flow_rate_nl_s: float
    """Flow rate of the injection in nanoliters per second."""

    start_time: datetime.datetime = datetime.datetime.now()
    """Time of the first injection, as a datetime object."""

    @pydantic.field_serializer("start_time", when_used="always")
    def serialize_start_time_field(self, start_time: datetime.datetime) -> str:
        return start_time.isoformat(sep=" ", timespec="seconds")

    is_anaesthetized: bool
    """Whether the subject was anaesthetized during the injection."""

    is_control: bool
    """Whether the purpose of the injection was a control."""

    notes: str | None = None
    """Text notes for the injection."""

    def to_json(self) -> dict[str, Any]:
        return self.model_dump()


@dataclasses.dataclass
class InjectionRecord:
    """A record of a set of injections.

    >>> i = Injection(
    ...     shield=npc_shields.shields.DR2002,
    ...     target_structure="VISp",
    ...     hemisphere="left",
    ...     depth_um=3000,
    ...     substance="Fluorogold",
    ...     manufacturer="Sigma",
    ...     identifier="12345",
    ...     total_volume_nl=1.0,
    ...     concentration_mg_ml=10.0,
    ...     flow_rate_nl_s=0.1,
    ...     start_time=datetime.datetime(2023, 1, 1, 12, 0),
    ...     fluorescence_nm=500,
    ...     notes="This was a test injection",
    ...     is_control=False,
    ...     is_anaesthetized=False,
    ... )
    >>> r = InjectionRecord(
    ...     injections=[i],
    ...     session="366122_20240101",
    ...     injection_day=1,
    ... )
    >>> r.to_json()
    {'injections': [{'shield': {'name': '2002', 'drawing_id': '0283-200-002'}, 'shield_hole': None, 'target_structure': 'VISp', 'hemisphere': 'left', 'depth_um': 3000.0, 'location_ap': None, 'location_ml': None, 'substance': 'Fluorogold', 'fluorescence_nm': 500, 'manufacturer': 'Sigma', 'identifier': '12345', 'total_volume_nl': 1.0, 'concentration_mg_ml': 10.0, 'flow_rate_nl_s': 0.1, 'start_time': '2023-01-01 12:00:00', 'is_anaesthetized': False, 'is_control': False, 'notes': 'This was a test injection'}], 'session': '366122_20240101', 'injection_day': 1}
    """

    injections: Iterable[npc_shields.types.Injection]
    """A record of each injection made."""

    session: str | npc_session.SessionRecord
    """Record of the session, including subject, date, session index."""

    injection_day: int
    """1-indexed day of experiment for the subject specified in `session`."""

    def to_json(self) -> dict[str, Any]:
        """Get a JSON-serializable representation of the injections."""
        return {
            "injections": [injection.to_json() for injection in self.injections],
            "session": self.session,
            "injection_day": self.injection_day,
        }


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations

from collections.abc import Iterator, MutableMapping
from typing import Any

import npc_session
from typing_extensions import Self

import npc_shields.shields
import npc_shields.types


class OccupiedHoleError(ValueError):
    pass


def validate_probe_insertion(
    shield: npc_shields.types.Shield,
    new_insertion: npc_shields.types.InsertionProbeMap,
    current_insertion: npc_shields.types.InsertionProbeMap | None = None,
) -> None:
    """
    >>> i = Insertion(npc_shields.shields.DR2002)
    >>> validate_probe_insertion(i.shield, {"A": "A1"}, i.probes)
    >>> validate_probe_insertion(i.shield, {"A": "A99"}, i.probes)
    Traceback (most recent call last):
    ...
    ValueError: probe='A' cannot be assigned to hole='A99': not a valid hole in insertion.shield.name='2002'
    """
    new_values = tuple(new_insertion.values())

    for probe, hole in new_insertion.items():
        if hole is None:
            continue
        if hole not in shield.labels:
            raise ValueError(
                f"{probe=} cannot be assigned to {hole=}: not a valid hole in {shield.name=}"
            )


def _get_default_probe_map() -> dict[str, str | None]:
    return dict.fromkeys("ABCDEF", None)


class ValidatedProbeMap(MutableMapping):
    """A dict-like container of probe letters mapped to insertion holes in a
    shield, with validation on assignment.

    - updating values validates that the holes are in the shield and not already
      occupied

    >>> i = ValidatedProbeMap(npc_shields.shields.DR2002)
    >>> i["A"] = "A1"
    >>> i["A"]
    'A1'
    """

    shield: npc_shields.types.Shield
    _probes: dict[str, str | None]

    def __init__(
        self,
        shield: npc_shields.types.Shield,
        probes: dict[str, str | None] | None = None,
    ) -> None:
        self.shield = shield
        self._probes = probes or _get_default_probe_map()
        for k, v in self._probes.items():
            self[k] = v

    def setmany(self, insertion: npc_shields.types.InsertionProbeMap) -> None:
        self.validate(insertion, with_current=False)
        self._probes.update(insertion)

    def validate(
        self,
        insertion: npc_shields.types.InsertionProbeMap,
        with_current: bool = True,
    ) -> None:
        validate_probe_insertion(
            shield=self.shield,
            new_insertion=insertion,
            current_insertion=self if with_current else None,
        )

    def __getitem__(self, key: str) -> str | None:
        return self._probes[key]

    def __setitem__(self, key: str, value: str | None) -> None:
        self.validate({key: value}, with_current=True)
        self._probes[key] = value

    def __iter__(self) -> Iterator[str]:
        return iter(self._probes)

    def __delitem__(self, key: str) -> None:
        raise NotImplementedError

    def __len__(self) -> int:
        return len(self._probes)


class Insertion:
    """A dict-like container of probe letters mapped to insertion holes in a
    shield, with other properties and methods fulfilling the `Insertion` protocol.

    >>> i = Insertion(npc_shields.shields.DR2002)
    """

    shield: npc_shields.types.Shield
    probes: npc_shields.types.InsertionProbeMap
    notes: dict[str, str | None]

    def __init__(
        self,
        shield: npc_shields.types.Shield,
        probes: dict[str, str | None] | None = None,
        notes: dict[str, str | None] | None = None,
    ) -> None:
        self.shield = shield
        self.probes = ValidatedProbeMap(shield, probes)
        self.notes = notes or _get_default_probe_map()

    def to_json(self) -> dict[str, Any]:
        """
        >>> i = Insertion(npc_shields.shields.DR2002)
        >>> j = i.to_json()
        >>> assert j["shield"]["name"] == "2002"
        """
        return {
            "shield": self.shield.to_json(),
            "probes": dict(self.probes),
            "notes": self.notes,
        }

    def to_svg(self) -> str:
        return npc_shields.shields.get_svg_data_with_insertions(
            self.shield, self.probes
        )

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Self:
        return cls(
            shield=npc_shields.shields.get_shield(data["shield"]["name"]),
            probes=data["probes"],
            notes=data["notes"],
        )


class InsertionRecord(Insertion):
    """A record of a set of probes inserted into a shield in a given session.

    >>> i = Insertion(npc_shields.shields.DR2002)
    >>> i.probes["A"] = "A1"
    >>> i.to_json()["probes"]["A"]
    'A1'
    """

    def __init__(
        self,
        shield: npc_shields.types.Shield,
        session: str | npc_session.SessionRecord,
        experiment_day: int,
        probes: dict[str, str | None] | None = None,
        notes: dict[str, str | None] | None = None,
    ) -> None:
        super().__init__(shield, probes, notes)
        self.session = npc_session.SessionRecord(session)
        self.experiment_day = experiment_day

    @property
    def session(self) -> npc_session.SessionRecord:
        """Record of the session (includes subject, date, session index)."""
        return self._session

    @session.setter
    def session(self, session: str | npc_session.SessionRecord) -> None:
        if getattr(self, "_session", None) is not None:
            raise AttributeError("session cannot be changed once set")
        self._session = npc_session.SessionRecord(session)

    @property
    def experiment_day(self) -> int:
        """1-indexed day of experiment for the subject specified in `session`."""
        return self._experiment_day

    @experiment_day.setter
    def experiment_day(self, experiment_day: int) -> None:
        if getattr(self, "_experiment_day", None) is not None:
            raise AttributeError("experiment_day cannot be changed once set")
        if not isinstance(experiment_day, int):
            raise TypeError(f"{experiment_day=} must be an integer")
        if experiment_day < 1:
            raise ValueError(
                f"{experiment_day=} must be a positive integer greater than 0"
            )
        self._experiment_day = experiment_day

    def to_json(self) -> dict[str, Any]:
        """
        >>> i = InsertionRecord(npc_shields.shields.DR2002, "366122_20240101", 1)
        >>> j = i.to_json()
        >>> assert j["session"] == "366122_20240101"
        """
        return {
            **super().to_json(),
            "session": self.session,
            "experiment_day": self.experiment_day,
        }

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Self:
        return cls(
            shield=npc_shields.shields.get_shield(data["shield"]["name"]),
            session=data["session"],
            experiment_day=data["experiment_day"],
            probes=data["probes"],
            notes=data["notes"],
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=(
            doctest.ELLIPSIS
            | doctest.NORMALIZE_WHITESPACE
            | doctest.FAIL_FAST
            | doctest.IGNORE_EXCEPTION_DETAIL
        )
    )
